#ifndef CMA_EXCEPTION_DEFINED
#define CMA_EXCEPTION_DEFINED

#include <stdlib.h>
#include <assert.h>
#include <string>

namespace CMa
{
	/**
	 * CMa base exception class.
	 *
	 * @author OM
	 */
	class Exception
	{
	public:
		virtual void getDescription(string& descr)=0;
	};

	/**
	 * CMa error class. When such error occurs, CMa remains in the state before
	 * the beginning of instruction that caused the error. Some errors are ment
	 * for debugging compliler and WILL BE REMOVED in the future.
	 *
	 * Possible errors:
	 *  - Stack errors
 	 *          - STACK_OVERFLOW
 	 *          - STACK_UNDERFLOW
 	 *
 	 *  - Aritmethical errors
 	 *	        - DIVISION_BY_ZERO
 	 *	        - MODULUS_IS_ZERO[TO BE REMOVED]
 	 *
 	 *  - Memory access errors
 	 *          - READ_FROM_NULL_POINTER
 	 *	        - WRITE_TO_NULL_POINTER
 	 *	        - READ_FROM_STACK_FRAME[TO BE REMOVED]
 	 *	        - WRITE_TO_STACK_FRAME[TO BE REMOVED] 			
 	 *	        - WRONG_HEAP_ADDRESS
 	 *	        - WRONG_HEAP_INTERVALL
 	 *
 	 *  - Code errors
 	 *	        - WRONG_CODE_ADDRESS
 	 *	        - WRONG_JUMP_ADDRESS[TO BE REMOVED]
 	 *	        - WRONG_CALL_ADDRESS[TO BE REMOVED]
 	 *	        - WRONG_OP_CODE
 	 *	        - WRONG_OP_ARGUMENT
 	 *	        - WRONG_OP_ARGUMENT_ADDRESS[TO BE REMOVED] 			
 	 *
 	 *  - Frame making errors
 	 *	        - CALL_UNDERFLOW [TO BE REMOVED]
 	 *	        - CORRUPTED_FRAME [TO BE REMOVED]
	 *
	 * @author OM, modified by SL
	 * STATUS OK
	 */
	
	class Error : public Exception
	{
	public:
		enum Code
		{
			//Stack errors
 			STACK_OVERFLOW,
			STACK_UNDERFLOW,
 			//Aritmethical errors
 			DIVISION_BY_ZERO,
			MODULUS_IS_ZERO,
			//Memory access errors
 			READ_FROM_NULL_POINTER,
 			WRITE_TO_NULL_POINTER,
 			READ_FROM_STACK_FRAME,
 			WRITE_TO_STACK_FRAME, 			
 			WRONG_HEAP_ADDRESS,
 			WRONG_HEAP_INTERVALL,
 			//Code errors
 			WRONG_CODE_ADDRESS,
 			WRONG_JUMP_ADDRESS,
 			WRONG_CALL_ADDRESS,
			WRONG_OP_CODE,
			WRONG_OP_ARGUMENT,
			WRONG_OP_ARGUMENT_ADDRESS, 			
 			//Frame making errors
		 	CALL_UNDERFLOW,
 			CORRUPTED_FRAME,
 			
 			OTHER
		};

		Error(Code code)
		{
			this->code = code;
		}

		Error(char* descr)
		{
			this->code = OTHER;
			this->descr = descr;
		}


		void getDescription(string& descr)
		{
			switch (code)
			{
			    //Stack errors
				case STACK_OVERFLOW:
					descr = "Stack overflow";				
				    break;
				case STACK_UNDERFLOW:
					descr = "Stack underflow";				
				    break;     			
 			    //Aritmethical errors
				case DIVISION_BY_ZERO:
                    descr = "Division by zero";
				    break;				
				case MODULUS_IS_ZERO:
				    descr = "Modulus is zero";
				    break;				
			    //Memory access errors
				case READ_FROM_NULL_POINTER:
				    descr = "Attempt to read from NULL-pointer";
				    break;				
				case WRITE_TO_NULL_POINTER:
				    descr = "Attempt to write to NULL-pointer";
				    break;				
				case READ_FROM_STACK_FRAME:
				    descr = "Unauthorized attempt to read data from stack frame";
				    break;				
				case WRITE_TO_STACK_FRAME:
				    descr = "Unauthorised attempt to write data to stack frame"; 			
				    break;				
				case WRONG_HEAP_ADDRESS:
				    descr = "Stack/heap address is invalid";
				    break;				
				case WRONG_HEAP_INTERVALL:
				    descr = "Stack/heap interval is invalid";
				    break;				
 			    //Code errors
				case WRONG_CODE_ADDRESS:
				    descr = "Instruction address is invalid";
                    break;
				case WRONG_JUMP_ADDRESS:
				    descr = "Jump address is invalid";
				    break;				
				case WRONG_CALL_ADDRESS:
				    descr = "Function call to invalid address";
				    break;				
				case WRONG_OP_CODE:
					descr = "Wrong operation code";				
				    break;				
				case WRONG_OP_ARGUMENT:
					descr = "Wrong operation argument";				
				    break;				
				case WRONG_OP_ARGUMENT_ADDRESS:
				    descr = "Address of instruction argument is invalid"; 		
				    break;					
 			    //Frame making errors
				case CALL_UNDERFLOW:
				    descr = "Error in CALL less arguments than needed";
				    break;				
				case CORRUPTED_FRAME:
				    descr = "Unable to jump back stack frame is corrupted";
				    break;				
                //Other errors
				case OTHER:
					descr = this->descr;
                    break;

				default:
					descr = "Unknown error";
			}
		}

	protected:
		Code code;
		string descr;
	};

	/**
	 * CMa interpreter's internal failure class.
	 *
	 * @author OM
	 */
	class Failure : public Exception
	{
	public:
		Failure(char* descr)
		{
			this->descr = descr;
		}

		virtual void getDescription(string& descr)
		{
			descr = this->descr;
		}

	protected:
		string descr;
	};
}

#endif

