#include <stdio.h>
#include <limits.h>

#include <vector>
#include <iostream>

#include "cma.h"
#include "arch_int.h"
#include "op.h"

using namespace CMa;
/**
 * A bit more elaborated control class. Allows to debug C-machine:
 *  - describes state before instruction
 *  - writes out instruction
 *  - executes one instruction
 *  - describes state after instruction and success/failure
 *
*/

class MyControl : public CMa::Control<Word_int>
{
    public:

     MyControl(Interpreter<
   		    Word_int, Int_int,
	    	Heap<Word_int, Int_int>,
		    FreeStore<Word_int, Int_int, Heap<Word_int, Int_int> >,
		    Code<Word_int, Int_int>
		    >* newinterpreter){ interpreter = newinterpreter;}
	 void onException(Exception* e)
	 {
		 string descr;
		 e->getDescription(descr);
		 cout << "STATUS: ";
		 cout << "Error: " << descr << endl;
	 }

	 void onComplete(Word_int result)
	 {
		cout << "Complete: " << Int_int(result).convertToint() << endl;
	 }

	 int onBreak()
	 {
	    static Int_int oldPC = (Int_int)-1;
	    Int_int PC;
	    Int_int SP;
	    int DSP;
	    int codeN;
	    int stackN;
	    Op::Code code;
	
	    //Result of last command
        if(oldPC>=(Int_int)0)
        {
            cout<<"FP="<<interpreter->FP.convertToint()<<" "
                <<"SP="<<interpreter->heap.getSP().convertToint()<<" "
                <<"EP="<<interpreter->heap.getEP().convertToint()<<" "
                <<"NP="<<interpreter->heap.getNP().convertToint()<<endl;
            if(interpreter->heap.verifyAddress(interpreter->heap.getSP()))
                cout<<"Top="<<interpreter->heap.top().convertToint() <<endl;
            cout<<"STATUS: OK"<<endl;
        }	
        PC = interpreter->PC;
        SP = interpreter->heap.getSP();
        oldPC =PC;

        if(!interpreter->code.verifyAddress(PC)) return 1;
        code = interpreter->code[PC].convertToCode();
        cout<<PC.convertToint()<<":";
        if(!op.getDescription(code,codeN, stackN,DSP))
        {
            cout<<"???"<<endl;
            return 1;
        }
        cout<<op.getName(code)<<" ";
        if(codeN>0 && !interpreter->code.verifyInterval(PC,PC+(Int_int)(codeN-1)))
        {
            cout<<"???"<<endl;
            return 1;
        }
        for(int i=1;i<codeN;++i) cout<<interpreter->code[PC+(Int_int)i].convertToint()<<",";
        cout<<"\b"<<endl;
        cout<<"FP="<<interpreter->FP.convertToint()<<" "
            <<"SP="<<interpreter->heap.getSP().convertToint()<<" "
            <<"EP="<<interpreter->heap.getEP().convertToint()<<" "
            <<"NP="<<interpreter->heap.getNP().convertToint()<<endl;

        cout<<"STACK: ";
        if(stackN>0 && !interpreter->heap.verifyInterval(SP-(Int_int)(stackN-1),SP))
        {
            cout<<"???"<<endl;
            return 1;
        }
        for(int i=0;i<stackN;++i) cout<<interpreter->heap.top(i).convertToint()<<",";
        cout<<"\b"<<endl;
		return 1;
    }
	
	Word_int onInput() throw (Failure)
	{
	    cout << "Input: ";
	    int value;
		cin >> value;
		return Word_int(value);
	}

	void onOutput_int(Word_int value) throw (Failure)
	{
	    cout << "Output: " << Int_int(value).convertToint() << endl;
	}
	void onOutput_char(Word_int value) throw (Failure)
	{
	   cout << "Output: " << (char)Int_int(value).convertToint() << endl;
	}
	
	private:

	/*Pointer to current C-machine */
	Interpreter<
		    Word_int, Int_int,
		    Heap<Word_int, Int_int>,
		    FreeStore<Word_int, Int_int, Heap<Word_int, Int_int> >,
		    Code<Word_int, Int_int>
			>* interpreter;
    Op op;

};


int main()
{
    try
    {
	Heap<Word_int, Int_int> heap;
	heap.setParams((Int_int)INT_MAX, 100000);
	heap.init();

	FreeStore< Word_int, Int_int, Heap<Word_int, Int_int> > freeStore(heap);
	freeStore.init();

	Code<Word_int, Int_int> code;
	code.data.resize(1000);
	code.data[0] = Word_int(Op::POP);
	code.data[1] = Word_int(Op::LOADC);
	code.data[2] = Word_int(Int_int(12345));
	code.init();

	// Test heap
	Int_int EP(10000);
	cout << heap.verifyEP(EP) << endl;
	heap.setEP(EP);
    cout << "EP is set." << endl;
	
	Int_int SP(5000);
	cout << heap.verifySP(SP) << endl;
	heap.setSP(SP);
    cout << "SP is set." << endl;

	
	// Test free store
	Int_int addr = (Int_int)freeStore.malloc((Int_int)5000);
	cout <<"new 5000=" << addr.convertToint() << endl;

	cout << heap.verifyAddress(addr + (Int_int)3000) << endl;
	heap[addr + (Int_int)3000] = Word_int(10);
    cout<<"Malloc seems valid."<<endl;
	
	heap.top() = Word_int(5000);
    cout<<"top()="<<(heap.top().value)<<endl;

	// Test code segment
	cout << code.verifyAddress((Int_int)0) << endl;
	cout << code.verifyInterval((Int_int)5, (Int_int)1000) << endl;

	cout << ((Int_int)code[(Int_int)2]).convertToint() << endl;

	code.deInit();
	freeStore.deInit();
	heap.deInit();
	cout << "Interpreter test"<<endl;


	// Test the interpreter
	Interpreter<
			Word_int, Int_int,
			Heap<Word_int, Int_int>,
			FreeStore<Word_int, Int_int, Heap<Word_int, Int_int> >,
			Code<Word_int, Int_int>
			> interpreter;

	interpreter.heap.setParams(Int_int(INT_MAX - CMA_MAX_SP_DELTA), 1000);

	interpreter.code.data.push_back(Word_int(Op::ENTER));
	interpreter.code.data.push_back(Word_int(Int_int(6)));
	interpreter.code.data.push_back(Word_int(Op::ALLOC));
	interpreter.code.data.push_back(Word_int(Int_int(1)));	
	interpreter.code.data.push_back(Word_int(Op::READI));
	interpreter.code.data.push_back(Word_int(Op::LOADC));
	interpreter.code.data.push_back(Word_int(Int_int(10)));
	interpreter.code.data.push_back(Word_int(Op::ADD));
	interpreter.code.data.push_back(Word_int(Op::HALT));

	MyControl myControl(&interpreter);
	interpreter.execute(&myControl);
	}
	catch(...)
	{
	    cout<<"Error happened"<<endl;
	}
}


