#ifndef CMA_OP_DEFINED
#define CMA_OP_DEFINED

#include <stdlib.h>

namespace CMa
{

/// Maximum increae of SP during execution of one operation
#define CMA_MAX_SP_DELTA 1

/// Maximum number of operation's arguments in the code segment
#define CMA_MAX_OP_CODE_ARG_COUNT 1

/// Maximum number of operation's arguments in the stack
#define CMA_MAX_OP_STACK_ARG_COUNT 2


	/**
 	 * Represents information about CMa operations.
	 *
	 * Internal tables are filled according Vesals' specification
	 *
	 * Described operations:
     *  -  Aritmetical operations
     *      - MINUS
     *      - INC
     *      - DEC
     *      - MUL
     *      - ADD
     *      - SUB
     *      - DIV
     *      - MOD
     *
     *  - Logical operations
     *      - NOT
     *      - AND
     *      - OR
     *
     *  - Comparison operators
     *      - EQ
     *      - NEQ
     *      - LE
     *      - LEQ
     *      - GE
     *      - GEQ
     *
     *  - Bitwise operators
     *      - BIT_NOT
     *      - BIT_AND
     *      - BIT_OR
     *      - BIT_XOR
     *      - BIT_SHL
     *      - BIT_SHR
     *
     *  - Various jumps
     *      - JUMP
     *      - JUMPZ
     *      - JUMPI
     *      - JUMPZS
     *      - JUMPNZ
     *
     *  - Store and load
     *      - LOADC
     *      - POP
     *      - LOAD
     *      - LOADA
     *      - STORE
     *      - STOREA
     *      - LOADRC
     *      - LOADR
     *      - STORER
     *
     *  - Duplication
     *      - DUP
     *      - MOVE
     *
     *  - Memory management
     *      - NEW
     *      - FREE
     *
     *  - Organisatory commands
     *      - MARK
     *      - CALL
     *      - ENTER
     *      - ALLOC
     *      - RETURN
     *      - HALT
     *
     *   - I/O commands
     *      - WRITEI
     *      - READI
     *      - WRITEC
	 *
	 * @author OM/Modified by SL
	 *
	 * STATUS: NOT TESTED/MAY CAUSE SERIOUS PROBLEMS
	 *
	 * TODO: Verify constructor.
 	 */
	class Op
	{
	    public:

		/// Possible CMa operation codes.
		enum Code
		{
			// Not a code
			WRONG_CODE = -1,
            //Aritmetical operations
            MINUS,
            INC,
            DEC,
            MUL,
            ADD,
            SUB,
            DIV,
            MOD,
			//Logical operations
            NOT,
            AND,
            OR,
			//Comparison operators
            EQ,
            NEQ,
            LE,
            LEQ,
            GE,
            GEQ,
			//Bitwise operators
            BIT_NOT,
            BIT_AND,
            BIT_OR,
            BIT_XOR,
            BIT_SHL,
            BIT_SHR,
			//Various jumps
            JUMP,
            JUMPZ,
            JUMPI,
            JUMPZS,
            JUMPNZ,
			//Store and load
            LOADC,
            POP,
            LOAD,
            LOADA,
            STORE,
            STOREA,
            LOADRC,
            LOADR,
            STORER,
			//Duplication
            DUP,
            MOVE,
			//Memory management
            NEW,
            FREE,
			//Organisatory commands
            MARK,
            CALL,
            ENTER,
            ALLOC,
            RETURN,
            HALT,
			//I/O commands
            WRITEI,
            READI,
            WRITEC,

			// Not a code
			MAX_CODE
		};
		
        /** Constructor fills instructions description table */
		Op()
		{
		    //Aritmetical operations
		    //describe(code, Arg, sArg, SD, name)
			describe(MINUS,  0,   1,    0,  "MINUS");
			describe(INC,    0,   1,    0,  "INC");
			describe(DEC,    0,   1,    0,  "DEC");
			describe(MUL,    0,   2,   -1,  "MUL");
			describe(ADD,    0,   2,   -1,  "ADD");
			describe(SUB,    0,   2,   -1,  "SUB");
			describe(DIV,    0,   2,   -1,  "DIV");
			describe(MOD,    0,   2,   -1,  "MOD");
			//Logical operations
			describe(NOT,    0,   1,    0,  "NOT");
			describe(AND,    0,   2,   -1,  "AND");
			describe(OR,     0,   2,   -1,  "OR");
			//Comparison operations
			describe(EQ,     0,   2,   -1,  "EQ");
			describe(NEQ,    0,   2,   -1,  "NEQ");
			describe(LE,     0,   2,   -1,  "LE");
			describe(LEQ,    0,   2,   -1,  "LEQ");
			describe(GE,     0,   2,   -1,  "GE");
			describe(GEQ,    0,   2,   -1,  "GEQ");
			//Bit operations
			describe(BIT_NOT,0,   1,    0,  "BIT_NOT");
			describe(BIT_AND,0,   2,   -1,  "BIT_AND");
			describe(BIT_OR, 0,   2,   -1,  "BIT_OR");
			describe(BIT_XOR,0,   2,   -1,  "BIT_XOR");
			describe(BIT_SHL,0,   2,   -1,  "BIT_SHL");
			describe(BIT_SHR,0,   2,   -1,  "BIT_SHR");
			//Jumping
			describe(JUMP,   1,   0,    0,  "JUMP");
			describe(JUMPZ,  1,   1,   -1,  "JUMPZ");
			describe(JUMPI,  1,   1,   -1,  "JUMPI");
			describe(JUMPZS, 1,   1,    0,  "JUMPZS");
			describe(JUMPNZ, 1,   1,    0,  "JUMPNZ");
            //Store and Load
			describe(LOADC,  1,   0,    1,  "LOADC");
			describe(POP,    0,   0,   -1,  "POP");
			describe(LOAD,   0,   1,    0,  "LOAD");
			describe(LOADA,  1,   0,    1,  "LOADA");
			describe(STORE,  0,   2,   -1,  "STORE");
			describe(STOREA, 1,   1,    0,  "STOREA");
			describe(LOADRC, 1,   0,    1,  "LOADRC");
			describe(LOADR,  1,   0,    1,  "LOADR");
			describe(STORER, 1,   1,    0,  "STORER");
			describe(DUP,    0,   1,    1,  "DUP");
			describe(MOVE,   1,   1,    0,  "MOVE");
			//(De)Allocation
			describe(NEW,    0,   1,    0,  "NEW");
			describe(FREE,   0,   1,   -1,  "FREE");
			//Organisatory commands
			describe(MARK,   0,   0,    4,  "MARK");
			describe(CALL,   1,   1,   -1,  "CALL");
			describe(ENTER,  1,   0,    0,  "ENTER");
			describe(ALLOC,  1,   0,    0,  "ALLOC");
			describe(RETURN, 0,   1,    0,  "RETURN");
			describe(HALT,   0,   0,    0,  "HALT");
			//I/O calls
			describe(WRITEI, 0,   1,   -1,  "WRITEI");
			describe(READI,  0,   0,    1,  "READI");
			describe(WRITEC, 1,   0,    0,  "WRITEC");
			//Table is unchecked -- simple copy from Vesal's specification.
		}

		/**
		 * Tells if the operation code is valid.
		 *
		 * @author OM
		 */
		inline bool isValid(Code code)
		{
			if ((code < 0)  || (code >= MAX_CODE)) return false;
			return opDescr[code].isValid;
		}

		/**
		 * Get description of the operation.
		 *
		 * @return true, if code was valid
		 *
		 * @author OM
		 */
		inline bool getDescription(
							Code code,
							int& codeArgN,
							int& stackArgN,
							int& SPdelta
							)
		{
			if ((code < 0)  || (code >= MAX_CODE)) return false;

			OpDescr& descr = opDescr[code];
			if (!descr.isValid) return false;

			codeArgN = descr.codeArgN;
			stackArgN = descr.stackArgN;
			SPdelta = descr.SPdelta;

			return true;
		}
		/**
		 * Returns name which corresponds to the code. Code must be valid
		 *
		 * @author OM
		 */

		inline char* getName(Code code)
		{
			assert(isValid(code));
			return opDescr[code].name;
		}

	private:

		/**
		 * A structure that describes one operation.
		 *
		 * @author OM
		 */
		struct OpDescr
		{
			/// Tells if this code is valid
			bool isValid;

			/// The number of arguments in the code segment
			int codeArgN;

			/// The number of arguments in the stack
			int stackArgN;

			/// The difference of SP before and after operation
			int SPdelta;

			/// Operation symbolic name
			char* name;

			OpDescr() : isValid(false), name(NULL)
			{}
		};

		OpDescr opDescr[ MAX_CODE ];


		/**
		 * Enter information into internal structures.
		 *
		 * @author OM
		 */
		void describe(
					Code code,
					int codeArgN,
					int stackArgN,
					int SPdelta,
					char* name
					)
		{
			opDescr[code].isValid = true;
			opDescr[code].codeArgN = codeArgN;
			opDescr[code].stackArgN = stackArgN;
			opDescr[code].SPdelta = SPdelta;
			opDescr[code].name = name;
		}

		/**
		 * Convenience method to describe functions.
		 *
		 * @author OM
		 */
		void describeFun(
					Code code,
					int stackArgN,
					char* name
					)
		{
			describe(code, 0, stackArgN, 1 - stackArgN, name);
		}
	};


	/**
	 * A static instance of class Op must be op(Op Op cause esoteric problems.
	 */
	static class Op op;
}

#endif

